#!/usr/bin/env python

# ########################################################################
#
# Copyright (C) 2010 by Marina Marcet Houben and Toni Gabaldon. All rights 
# reserved.  
# email: mmarcet[at]crg.es
#
# TreeKO is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#  
# TreeKO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# ########################################################################


import ete2
import subtree_reconstruction as ms
import sys
import re
import os
import os.path
from optparse import OptionParser


###########################################################################################################
# This program calculates the phylome support for a species tree. For each node in the species tree
# alternative topologies will be suggested (3 or 15 depending on the number of grand-children the node has)
# These alternative topologies will be searched for in a list of trees (can be a phylome) and the percentage
# of tree supporting each alternative is calculated. The program produces two files, the first one will
# have the statistics of each alternative node, the second will be the tree with the phylome_support represented
# at the branches, bootstrap or alrt support will be shown in brakets next to it.
# Topologies are represented as lists of species names, for instance: Sce-Spa|Smi Sku.
# Species separated by a "-" belong to the same group, groups are separated by either a pipe "|" or a tab.
# Pipes separate sister groups while tabs separate scaled groups.
###########################################################################################################

###########################################################################################################
# Parser
###########################################################################################################


def load_initial_data(d):
	if d == "g":
		info = {"tree_mode" : "s", "species_position" : "1", "orto_mode" : "s", "species2age" : {}, "species_tree" : "", "root_species" : "", "limit" : 0, "overlap" : False}
	else:
		info = {"kind_root" : "m", "spec_filter" : None, "prot_filter" : None, "bootstrap" : None}
	return info
		
def parse_conf_file(file):
	
	#Initialize data:
	general_info = load_initial_data("g")
	info_tree1 = load_initial_data("t")
	info_tree2 = load_initial_data("t")
	species_limit = "3"
	verbose = False
	for line in open(file):
		line = line.strip()
		if line != "":
			if line[0] != "#":
				if " " in line:
					dad = line.split(" ")
				else:
					dad = line.split("\t")
				if dad[0].lower() == "root_method1":
					kind_root = dad[1].lower()
					if kind_root != "f" and kind_root != "s" and kind_root != "m" and kind_root != "n" and kind_root != "fi" and kind_root != "d":
						kind_root = ""
						sys.exit("The root method provided is not supported, please check the manual for the proper configuration")
					info_tree1["kind_root"] = kind_root
				elif dad[0].lower() == "root_method2":
					kind_root = dad[1].lower()
					if kind_root != "f" and kind_root != "s" and kind_root != "m" and kind_root != "n" and kind_root != "fi" and kind_root != "d":
						kind_root = ""
						sys.exit("The root method provided is not supported, please check the manual for the proper configuration")
					info_tree2["kind_root"] = kind_root
				elif dad[0].lower() == "species_position":
					general_info["species_position"] = dad[1].lower()
				elif dad[0].lower() == "species_limit":
					species_limit = dad[1].lower()
					if species_limit != "1" and species_limit != "2" and species_limit != "3" and species_limit != "4" and species_limit != "5" and species_limit != "6" and species_limit != "7" and species_limit != "_" and species_limit != "-" and species_limit != "#" and species_limit != ":" and species_limit != "." and species_limit != "all" and species_limit != "phylomedb":
						sys.exit("The species delimiter is not accepted. Please, check the manual for accepted limits for the species name")
				elif dad[0].lower() == "root_data":
					if ":" in dad[1]:
						general_info["species2age"][dad[1].split(":")[0]] = int(dad[1].split(":")[1])
					else:
						sys.exit("Information should be provided in the following format:\nspecies_name:level")
				elif dad[0].lower() == "root_species":
					general_info["root_species"] = dad[1]
				elif dad[0].lower() == "orto_mode":
					orto_mode = dad[1].lower()
					if orto_mode != "s" and orto_mode != "r" and orto_mode != "m":
						sys.exit("The orthology detection mode is not valid, please revise the information provided by orto_mode")
					general_info["orto_mode"] = orto_mode

				elif dad[0].lower() == "species_tree":
					general_info["species_tree"] = dad[1]
					
				elif dad[0].lower() == "support1":
					try:
						info_tree1["bootstrap"] = float(dad[1])
					except:
						sys.exit("There was a problem with the support value for tree 1, please check your configuration file")
				elif dad[0].lower() == "support2":
					try:
						info_tree2["bootstrap"] = float(dad[1])
					except:
						sys.exit("There was a problem with the support value for tree 2, please check your configuration file")

				elif dad[0].lower() == "limit":
					try:
						general_info["limit"] = int(dad[1])
					except:
						print "Limit has to have a numerical value, put 0 if you do not whish to use any limit"
						exit()
				elif dad[0].lower() == "verbose":
					verbose = True
				else:
					print "WARNING: The information provided in the configuration file is not in available for this program, please check the manual for further information."
					print line
	if general_info["orto_mode"] == "r":
		if general_info["species_tree"] != "":
			if not os.path.exists(general_info["species_tree"]):
				sys.exit("Unable to find the species tree, please check your configuration file")
		else:
			sys.exit("A species tree needs to be provided if reconciliation is chosen as the orthology prediction program, please provide the path to the species tree in the configuration file")
		if info_tree2["bootstrap"] != None or info_tree1["bootstrap"] != None:
			sys.exit("At the moment multifurcated trees cannot be treated with tree reconciliation methods, please switch to species overlap or remove the support limits")
			exit()
	if info_tree1["kind_root"] != "n":
		if info_tree1["kind_root"] == "":
			info_tree1["kind_root"] = "m"
		elif info_tree1["kind_root"] == "f" or info_tree1["kind_root"] == "fi":
			if general_info["species2age"] == {}:
				sys.exit("The information for rooting at the farthest node is not provided, please check the manual for further information")
		elif info_tree1["kind_root"] == "s":
			if general_info["root_species"] == "":
				sys.exit("The information needed to root at a given protein is not provided, please check the manual for further information")
				exit()
		elif info_tree1["kind_root"] == "n" and not is_rooted1:
			sys.exit("Tree1 is not rooted, please provide a rooted tree or change your configuration file")
		if info_tree2["kind_root"] != "n":
			if info_tree2["kind_root"] == "":
				info_tree2["kind_root"] = "m"
			elif info_tree2["kind_root"] == "f" or info_tree2["kind_root"] == "fi":
				if general_info["species2age"] == {}:
					sys.exit("The information for rooting at the farthest node is not provided, please check the manual for further information")
			elif info_tree2["kind_root"] == "s":
				if general_info["root_species"] == "":
					sys.exit("The information needed to root at a given protein is not provided, please check the manual for further information")
			elif info_tree2["kind_root"] == "n" and not is_rooted2:
				sys.exit("Tree2 is not rooted, please provide a rooted tree or change your configuration file")
	if general_info["species_position"] != "1" and general_info["species_position"] != "-1":
		sys.exit("The species position has to be either 1 or -1, please check your configuration file")
	if general_info["tree_mode"] == "":
		general_info["tree_mode"] = "s"
	if species_limit != "3":
		general_info["tree_mode"] = "s"+species_limit
	return general_info,info_tree1,info_tree2,verbose

###########################################################################################################
# Get toplogies from species tree. CAREFUL: Tree must be apropriately rooted!
###########################################################################################################

def obtain_topo_from_node(node):
	#This get the children/grandchildren for a given node
	topo = []
	for child in node.get_children():
		if child.get_children() == []:
			topo.append(child.get_leaf_names()[0])
		else:
			for grandchild in child.get_children():
				string = "-".join(grandchild.get_leaf_names())
				topo.append(string)
	return topo

def combine3(children):
	#Makes the alternative combinations when the number of grandchildren is 3
	alternative_topos = {}
	alternative_topos[0] = [children[0]+"|"+children[1],children[2]]
	alternative_topos[1] = [children[0]+"|"+children[2],children[1]]
	alternative_topos[2] = [children[1]+"|"+children[2],children[0]]
	return alternative_topos

def combine4(children):
	#Makes the alternative combinations when the number of grandchildren is 4
	alternative_topos = {}
	alternative_topos[0] = [children[0]+"|"+children[1],children[2],children[3]]
	alternative_topos[1] = [children[0]+"|"+children[1],children[3],children[2]]
	alternative_topos[2] = [children[0]+"|"+children[2],children[1],children[3]]
	alternative_topos[3] = [children[0]+"|"+children[2],children[3],children[1]]
	alternative_topos[4] = [children[0]+"|"+children[3],children[1],children[2]]
	alternative_topos[5] = [children[0]+"|"+children[3],children[2],children[1]]
	alternative_topos[6] = [children[1]+"|"+children[2],children[0],children[3]]
	alternative_topos[7] = [children[1]+"|"+children[2],children[3],children[0]]
	alternative_topos[8] = [children[1]+"|"+children[3],children[0],children[2]]
	alternative_topos[9] = [children[1]+"|"+children[3],children[2],children[0]]
	alternative_topos[10] = [children[2]+"|"+children[3],children[0],children[1]]
	alternative_topos[11] = [children[2]+"|"+children[3],children[1],children[0]]
	alternative_topos[12] = [children[0]+"|"+children[1],children[2]+"|"+children[3]]
	alternative_topos[13] = [children[0]+"|"+children[2],children[1]+"|"+children[3]]
	alternative_topos[14] = [children[0]+"|"+children[3],children[1]+"|"+children[2]]
	return alternative_topos
	
def make_alternative_topologies(children):
	#directs the creation of alternative topologies
	if len(children) == 3:
		alternative_topos = combine3(children)
	elif len(children) == 4:
		alternative_topos = combine4(children)
	else:
		print "Something went wrong with the children"
		exit()
	return alternative_topos
	
def obtain_topologies(species_tree):
	#Obtains all the alternative topologies for all the nodes that are candidates to obtain the phylome support
	all_topologies = {}
	for node in species_tree.traverse():
		if node.is_leaf() or node.is_root():
			pass
		else:
			#print len(node.get_leaves()),len(species_tree.get_leaves()),node.get_leaf_names()
			if len(node.get_leaves()) <= 2 or len(node.get_leaves()) == len(species_tree.get_leaves()):
				pass
			else:
				children = obtain_topo_from_node(node)
				topos = make_alternative_topologies(children)
				key = "-".join(children)
				all_topologies[key] = topos
	return all_topologies


###########################################################################################################
# Tree modification part
###########################################################################################################



#file = file name for tree
#kroot = kind of rooting: midpoint(m), spe2age(f), species (s)
#spe2age = dictionary if f rooting is chosen
#tree_mode = trees should be treated as protein trees or species should be identified
#root_species = species where root should be placed if species rooting mode chosen
#check_species = bool, False if there could be repeated name tags at leaves
#limit = integer, indicates whether a limit for the number of subtrees should be enabled

#Obtains a list of subtrees for each tree to be considered, only subtrees that contain the seed will be used
def get_subtrees(file_name,general_info,info_tree2,check_species,only_orthologs,seed,list_subtrees):
	colection,total_tree_length,deli,stop = ms.make_subtrees(file_name,general_info,info_tree2,check_species)
	col = []
	for tree in colection:
		if type(tree) == str:
			col.append(tree)
	collection = []
	if only_orthologs and not seed == None:
		for nw in col:
			try:
				tree = ete2.PhyloTree(nw)
				if seed in tree.get_leaf_names():
					collection.append(nw)
			except:
				pass
	else:
		for nw in col:
			try:
				tree = ete2.PhyloTree(nw)
				collection.append(nw)
			except:
				pass
	if seed:
		list_subtrees[seed] = collection
	else:
		code = "No_seed_"+str(len(list_subtrees))
		list_subtrees[code] = collection
	return list_subtrees

#Checks wether there are representatives of each part of the topology in the tree
def check_species_presence(nw,topo,general_info):
	tree = ms.load_tree(nw,general_info["tree_mode"],general_info["species_position"],0)
	species = tree.get_species()
	acceptable = True
	for topo1 in topo:
		for t in topo1.split("|"): 
			trobat_topo = False
			for s in t.split("-"):
				if s in species:
					trobat_topo = True
			if not trobat_topo:
				acceptable = False
	trobat_topo = False
	for s in species:
		for t in topo:
			if s in t:
				trobat_topo = True
	if not trobat_topo:
		acceptable = False
	return acceptable,tree

# Function used to prune the subtrees so that only species present in a given topology are considered, 
# it also adds the group feature that will determine to which topological group each leaf belongs to
def prune_trees(tree,topo):
	species = []
	i = 0
	j = 0
	rename = {}
	for t in topo:
		i = i + 1
		j = 0
		for d in t.split("|"):
			j = j + 1
			for s in d.split("-"):
				species.append(s)
				rename[s] = str(i)+str(j)
	leaves_to_keep = []
	for leaf in tree.get_leaves():
		if leaf.species in species:
			leaves_to_keep.append(leaf.name)
	tree.prune(leaves_to_keep)
	for leaf in tree.get_leaves():
		leaf.add_features(group=rename[leaf.species])
	#tree.show()
	return tree

# Sorts the subtrees so that only those that have all members of the topo will be used, 
# also it renames the leaves so that they addopt their group name to simplify further analyses
def check_presence_trees(list_trees,tree,list_trees_renamed,seed):
	trobat = False
	for nw in list_trees:
		t = ete2.PhyloTree(nw)
		intersection = set(t.get_leaf_names()).intersection(set(tree.get_leaf_names()))
		if len(intersection) == len(t.get_leaf_names()) and len(intersection) == len(tree.get_leaf_names()):
			trobat = True
	if not trobat:
		list_trees.append(tree.write())
		for leaf in tree.get_leaves():
			leaf.name = leaf.group
		list_trees_renamed.append(tree.write())
	return list_trees,list_trees_renamed

# Deletes monophyletic groups of species that belong to the same group so that only one member is left.
def delete_intern_duplicates(tree):
	clean = False
	while not clean:
		clean = True
		for node in tree.traverse():
			if len(node.get_leaf_names()) == 2 and not node.is_root():
				if len(set(node.get_leaf_names())) == 1:
					clean = False
					child = node.get_children()[0]
					child.delete()
	return tree

# Deletes duplications involving more than one group
def delete_intern_groups(tree,topo):
	tree = delete_intern_duplicates(tree)
	if "|" not in topo[0] or "|" not in topo[1]:
		i = 2
		while i <= len(topo):
			for node in tree.traverse():
				if len(node.get_leaves()) == i and not node.is_root() and not node.up.is_root():
					sist = node.get_sisters()[0]
					#print node.get_sisters()[0].get_leaf_names(),node.get_leaf_names()
					intersect = set(sist.get_leaf_names()).intersection(set(node.get_leaf_names()))
					#print intersect
					if len(intersect) == len(set(sist.get_leaf_names())) and len(intersect) == len(set(node.get_leaf_names())):
						#print "DELETE"
						n = node.up
						if n.name != "DELETE":
							n.name = "DELETE"
			i = i + 1
		for node in tree.traverse():
			if node.name == "DELETE":
				child = node.get_children()[0]
				child.detach()
				node.delete()
	return tree

###########################################################################################################
# Phylome support per node
###########################################################################################################

def get_phylome_support_per_node(topo,list_subtrees,general_info):
	aproved_total, total_trees = 0.0,0
	for seed,subtrees in list_subtrees.iteritems():
		percentage,assessed = check_topologies(subtrees,topo,seed,general_info)
		if assessed:
			aproved_total = aproved_total + percentage
			total_trees = total_trees + 1
	return aproved_total,total_trees

#Calculates how many subtrees contain a given topology
def check_topologies(subtrees,topo,seed,general_info):
	final_tree_list,list_trees_renamed,correct,assessed = [],[],False,True
	supporting = 0
	candidates = 0
	percentage = 0
	for nw in subtrees:
		acceptable,tree = check_species_presence(nw,topo,general_info)
		if acceptable:
			tree = prune_trees(tree,topo)
			final_tree_list,list_trees_renamed = check_presence_trees(final_tree_list,tree,list_trees_renamed,seed)
	for nw in list_trees_renamed:
		tree = delete_intern_groups(ete2.PhyloTree(nw),topo)
		#tree.show()
		correct = check_topo(tree,topo)
		candidates = candidates + 1
		if correct:
			supporting = supporting + 1
	if candidates != 0:
		percentage = float(supporting) / float(candidates)
	else:
		assessed = False
	return percentage, assessed

# Checks whether the topology is within the tree or not
def check_topo(tree,topo):
	trobat = True
	if "|" not in topo[0] or "|" not in topo[1]:
		for node in tree.traverse():
			if len(node.get_leaves()) == 2:
				leaves = node.get_leaf_names()
				if leaves[0][0] != leaves[1][0]:
					trobat = False
				#print "Node inicial",trobat
				if trobat:
					i = 1
					while i < len(topo):
						try:
							sis = node.get_sisters()[0]
							#print sis.get_leaf_names()
							num = ""
							correct = True
							for leaf in sis.get_leaf_names():
								if num == "":
									num = leaf[0]
								else:
									if leaf[0] != num:
										correct = False
						except:
							 correct = False
						i = i + 1
						#print "Nodes interns",correct,num
						if correct and int(num) == i:
							pass
						else:
							trobat = False
						node = node.up
	else:
		trobat1 = False
		trobat2 = False
		for node in tree.traverse():
			if len(node.get_leaves()) == 2:
				leaves = node.get_leaf_names()
				if leaves[0][0] == leaves[1][0]:
					if leaves[0][0] == "1":
						trobat1 = True
					elif leaves[0][0] == "2":
						trobat2 = True
		if trobat1 and trobat2:
			trobat = True
		else:
			trobat = False
	return trobat




###########################################################################################################
# Read inputs
###########################################################################################################


def get_list_subtrees(subtree_file,only_orthologs,verbose,info_tree2,general_info):
	list_subtrees = {}
	check_species = True
	if verbose:
		total = 0
		for line in open(subtree_file):
			total = total+1
		i = 0
	for line in open(subtree_file):
		line = line.strip()
		if " " in line:
			files = line.split(" ")
		else:
			files = line.split("\t")
		if len(files) == 2:
			if files[0] in files[1]:
				seed = files[0]
			else:
				seed = None
			if "(" in files[1] and ")" in files[1] and ";" in files[1]:
				file = open("tmp_tree.txt","w")
				print >>file,files[1]
				file.close()
				file_name = "tmp_tree.txt"
			else:
				file_name = files[1]
		else:
			if "(" in line and ")" in line and ";" in line:
				file = open("tmp_tree.txt","w")
				print >>file, line
				file.close()
				file_name = "tmp_tree.txt"
			else:
				file_name = line
			seed = None
#		if only_orthologs and seed==None:
#			sys.exit("Pruned trees cannot be filtered without a seed, please change your tree list formating by adding a seed at the beginning")
		list_subtrees = get_subtrees(file_name,general_info,info_tree2,check_species,only_orthologs,seed,list_subtrees)
		if verbose:
			i = i + 1
			if i%100 == 0:
				print str(i)+" out of "+str(total)
	if os.path.exists("tmp_tree.txt"):
		cmd = "rm tmp_tree.txt"
		resu = os.popen(cmd)
	return list_subtrees

def locate_node(list_spe,tree):
	species = list_spe.split("-")
	spe1 = tree.get_leaves_by_name(species[0])[0]
	spe2 = tree.get_leaves_by_name(species[1])[0]
	common = spe1.get_common_ancestor(spe2)
	while len(common.get_leaves()) != len(species) and not common.is_root():
		common = common.up
	return common



###########################################################################################################
# Create outfiles
###########################################################################################################


#Create the stats file so that all the alternative topologies have the percentage of trees that support them
def calculate_support_stats(all_topologies,support,outstats):
	percent_support = {}
	for t in all_topologies:
		print >>outstats,"Results for the partition containing:\t"+", ".join(t.split("-"))
		percent_support[t] = {}
		total = 0
		for key,val in support[t].iteritems():
			total = total + val
		for key,val in support[t].iteritems():
			if total != 0:
				percent = val/total
			else:
				percent = 0
			percent_support[t][key] = percent
			print >>outstats,"\t".join(all_topologies[t][key])+"\t%.3f\t%.3f" %(val,percent)
		print >>outstats,"\n"
	outstats.close()
	return percent_support

# Looks whether two species apear in sister nodes in a given tree
def is_sister(tree,spe1,spe2):
	node = tree.get_leaves_by_name(spe1)[0]
	sis = node.get_sisters()[0]
	if spe2 in sis.get_leaf_names():
		s = True
	else:
		s = False
	return s

# Prints the species tree with the node support, branch distances are ommited
def assign_support_to_tree(all_topologies,support,tree,outtree,outnw):
	for node in tree.traverse():
		node.dist = 1.0
	for t in all_topologies:
		correct_topo = ""
		common = locate_node(t,tree)
		common_nw = common.write()
		if len(all_topologies[t]) == 15:
			for key, topo in all_topologies[t].iteritems():
				if "|" in topo[0] and "|" in topo[1]:
					spe1 = topo[0].split("|")[0].split("-")[0]
					spe2 = topo[0].split("|")[1].split("-")[0]
					spe3 = topo[1].split("|")[0].split("-")[0]
					spe4 = topo[1].split("|")[1].split("-")[0]
					to_keep = [spe1,spe2,spe3,spe4]
					common = ete2.PhyloTree(common_nw)
					common.prune(to_keep)
					if is_sister(common,spe1,spe2) and is_sister(common,spe3,spe4):
						correct_topo = key
		else:
			for key, topo in all_topologies[t].iteritems():
				spe1 = topo[0].split("|")[0].split("-")[0]
				spe2 = topo[0].split("|")[1].split("-")[0]
				spe3 = topo[1].split("|")[0].split("-")[0]
				to_keep = [spe1,spe2,spe3]
				common = ete2.PhyloTree(common_nw)
				common.prune(to_keep)
				if is_sister(common,spe1,spe2):
					correct_topo = key
				
					
		common = locate_node(t,tree)
		common.dist = support[t][correct_topo]
	IP = ete2.TreeImageProperties()
	IP.draw_branch_length = True
	IP.force_topology = True
	tree.write(outfile=outnw)
	tree.render(outtree,img_properties=IP)
	
	
def phylome_support(species_tree_file,tree_file,conf_file,outstats,outtree,outnw):
	#Parse options
	if conf_file:
		general_info, info_tree1, info_tree2, verbose = parse_conf_file(conf_file)
	else:
		general_info = load_initial_data("g")
		info_tree1 = load_initial_data("t")
		info_tree2 = load_initial_data("t")
		species_limit = "3"
		verbose = True
	
	#SPECIES TREE INFILE
	if not os.path.exists(species_tree_file):
		sys.exit("Unable to find the species tree file, please, check your command line")
	else:
		t = ete2.PhyloTree(species_tree_file)
		if len(t.get_children()) == 3:
			sys.exit("Species tree needs to be rooted, else unexpected partitions may be computed")
		else:
			all_topologies = obtain_topologies(ete2.PhyloTree(species_tree_file))
	
	#LISTS OF TREES
	if verbose:
		print "Getting pruned trees"
	#if options.orto:
	only_orthologs = True
	#else:
		#only_orthologs = False
	if tree_file:
		if not os.path.exists(tree_file):
			sys.exit("The file containing the list of trees was not found, please check your command line")
		else:
			list_subtrees = get_list_subtrees(tree_file,only_orthologs,verbose,info_tree2,general_info)
	if len(list_subtrees) == 0:
		sys.exit("No trees were found, please check your input data. If a phylomeID was provided, check whether this is available to the public user")
	if verbose:
		print "Getting support"

	support = {}
	for t in all_topologies:
		total = 0.0
		support[t] = {}
		for key,topo in all_topologies[t].iteritems():
			aproved_total,total_trees = get_phylome_support_per_node(topo,list_subtrees,general_info)
			support[t][key] = aproved_total
	support = calculate_support_stats(all_topologies,support,outstats)
	tree = assign_support_to_tree(all_topologies,support,ete2.PhyloTree(species_tree_file),outtree,outnw)


